#include "AnalysisReport.h"
#include "Parser.h"
#include <iomanip>
using namespace std;

std::string AnalysisReport::name() const
{
	return string("Analysis Report");
}

int AnalysisReport::print(const STVec & trees, std::ostream & os)
{
	modules.clear();
	functions.clear();
	funcs.clear();
	sdata.clear();
	gdata.clear();
	callgraph.clear();
	data_access.clear();

	collect_defined_functions(trees);
	collect_data(trees);
	collect_function_info(trees);

	os << setiosflags(ios::left);
	print_modules(os);
	print_defined_functions(os);
	print_calls(os);
	print_static_data(os);
	print_global_data(os);
	print_data_accesses(os);

	modules.clear();
	functions.clear();
	funcs.clear();
	sdata.clear();
	gdata.clear();
	callgraph.clear();
	data_access.clear();

	return 0;
}

void AnalysisReport::print_defined_functions(std::ostream & os) const
{
	for (DSet::const_iterator it = funcs.begin(); it != funcs.end(); ++it)
		os << "F" << "\t" << "\"" << *it << "\"" << endl;
}

void AnalysisReport::print_calls(std::ostream & os) const
{
	for (DDMMap::const_iterator it = callgraph.begin(); it != callgraph.end(); ++it)
		os << "C" << "\t" << "\"" << it->first << "\"" << "\t" << "\"" << it->second << "\"" << endl;
}

void AnalysisReport::print_static_data(std::ostream & os) const
{
	for (DSet::const_iterator it = sdata.begin(); it != sdata.end(); ++it)
		os << "S" << "\t" << "\"" << *it<< "\"" << endl;
}

void AnalysisReport::print_global_data(ostream & os) const
{
	for (DSet::const_iterator it = gdata.begin(); it != gdata.end(); ++it)
		os << "G" << "\t" << "\"" << *it<< "\"" << endl;
}

void AnalysisReport::print_data_accesses(std::ostream & os) const
{
	for (DDMMap::const_iterator it = data_access.begin(); it != data_access.end(); ++it)
		os << "A" << "\t" << "\"" << it->first << "\"" << "\t" << "\"" << it->second << "\"" << endl;
}

void AnalysisReport::print_modules(std::ostream & os) const
{
	for (SSet::const_iterator it = modules.begin(); it != modules.end(); ++it)
		os << "M" << "\t" << "\"" << *it << "\"" << endl;
}


bool AnalysisReport::is_function_defined_in_module(const std::string & function, const std::string & module) const
{
	if (module.size() <= 0) return false;
	for (SSMMap::const_iterator func = functions.lower_bound(module); func != functions.upper_bound(module); ++func)
		if (func->second == function) return true;
	return false;
}

std::string AnalysisReport::get_function_name(const Node * function_definition) const
{
	if (!function_definition->is_type(Parser::TOKEN_FUNCTION_DEFINITION)) return string("");
	for (Node::const_iterator i = function_definition->begin(); i != function_definition->end(); ++i)
	{
		Node * node = *i;
		if (node->is_type(Parser::TOKEN_DIRECT_DECLARATOR)) return (*node)[0]->identifier();
		else if (node->is_type(Parser::TOKEN_DECLARATOR))
		{
			for (Node::const_iterator j = node->begin(); j != node->end(); ++j)
			{
				Node * n = *j;
				if (n->is_type(Parser::TOKEN_DIRECT_DECLARATOR)) return (*n)[0]->identifier();
			}
		}
	}
	return string("");
}

std::string AnalysisReport::find_module_of_function(const std::string & function) const
{
	for (SSMMap::const_iterator it = functions.begin(); it != functions.end(); ++it)
		if (it->second == function) return it->first;
	return string("");
}

void AnalysisReport::collect_defined_functions(const STVec & trees)
{
	for (STVec::const_iterator tree = trees.begin(); tree != trees.end(); ++tree)
	{
		for (SyntaxTree::const_iterator node = (*tree)->begin(); node != (*tree)->end(); ++node)
		{
			if ((*node)->is_type(Parser::TOKEN_FUNCTION_DEFINITION))
			{
				string function_name = get_function_name(*node);
				string module_name = (*node)->get_module();

				modules.insert(module_name);
				functions.insert(SSMMap::value_type(module_name, function_name));
				funcs.insert(Declaration(function_name, module_name));
			}
		}
	}
}

void AnalysisReport::collect_function_info(const STVec & trees)
{
	for (STVec::const_iterator tree = trees.begin(); tree != trees.end(); ++tree)
	{
		for (SyntaxTree::const_iterator node = (*tree)->begin(); node != (*tree)->end(); ++node)
		{
			if ((*node)->is_type(Parser::TOKEN_FUNCTION_DEFINITION))
			{
				caller.set_name(get_function_name(*node));
				caller.set_module((*node)->get_module());
				collect_function_calls(*node);
				collect_data_accesses_from_func(*node);
			}
		}
	}
}

void AnalysisReport::collect_function_calls(Node * node)
{
	for (Node::const_iterator i = node->begin(); i != node->end(); ++i)
	{
		if ((*i)->is_type(Parser::TOKEN_POSTFIX_EXPRESSION))
		{
			if ((**i)[0]->is_type(Parser::TOKEN_IDENTIFIER)
				&& (**i)[1]->is_type(Parser::TOKEN_PARENTHESIS_LEFT))
			{
				Declaration callee((**i)[0]->identifier(), "");
				if (is_function_defined_in_module(callee.get_name(), caller.get_module())) callee.set_module(caller.get_module());
				else callee.set_module(find_module_of_function(callee.get_name()));
				bool found = false;
				for (DDMMap::const_iterator i = callgraph.lower_bound(caller); !found && i != callgraph.upper_bound(caller); ++i)
				{
					found = (i->second == callee);
				}
				if (!found)
				{
					callgraph.insert(DDMMap::value_type(caller, callee));
					if (callee.get_module().size() <= 0)
					{
						cerr << __FUNCTION__ << ":" << __LINE__ << ": ERROR: Callgraph: callee '"
							<< callee.get_name() << "' with unknown module" << endl;
					}
				}
			}
		}
		collect_function_calls(*i);
	}
}

void AnalysisReport::extract_identifiers_from_declaration(const Node * decl, DataCollectorContext & ctx)
{
	if (!decl) return;
	if (!decl->is_type(Parser::TOKEN_DECLARATION)) return;
	const Node * sub = (*decl)[0];
	if (sub->is_type(Parser::TOKEN_DECLARATION_SPECIFIERS))
	{
		for (Node::const_iterator i = sub->begin(); i != sub->end(); ++i)
		{
			switch ((*i)->get_type())
			{
				case Parser::TOKEN_EXTERN:  return; // external symbols are not of interest
				case Parser::TOKEN_TYPEDEF: return; // typedefs are not of interest
			}
		}
	}
	sub = (*decl)[1];
	switch (sub->get_type())
	{
		case Parser::TOKEN_IDENTIFIER:
			ctx.add(sub->identifier());
			break;
		case Parser::TOKEN_INIT_DECLARATOR:
			ctx.add((*sub)[0]->identifier());
			break;
		case Parser::TOKEN_INIT_DECLARATOR_LIST:
			for (Node::const_iterator i = sub->begin(); i != sub->end(); ++i)
				if ((*i)->is_type(Parser::TOKEN_IDENTIFIER))
					ctx.add((*i)->identifier());
			break;
		case Parser::TOKEN_DIRECT_DECLARATOR:
			string s;
			if (extract_identifier_of_direct_declarator(sub, s)) ctx.add(s);
			break;
	}
}

void AnalysisReport::collect_unique_data_access(const Declaration & data)
{
	for (DDMMap::const_iterator i = data_access.lower_bound(caller); i != data_access.upper_bound(caller); ++i)
		if (i->second == data) return;
	data_access.insert(DDMMap::value_type(caller, data));
}

void AnalysisReport::collect_data_recursive_depth(const Node * node, DataCollectorContext & ctx)
{
	if (!node) return;
	for (Node::const_iterator i = node->begin(); i != node->end(); ++i)
	{
		switch ((*i)->get_type())
		{
			case Parser::TOKEN_BRACE_LEFT: // '{'
				ctx.enter_scope();
				break;
			case Parser::TOKEN_BRACE_RIGHT: // '}'
				ctx.leave_scope();
				break;
			case Parser::TOKEN_DECLARATION:
				extract_identifiers_from_declaration(*i, ctx);
				break;
			case Parser::TOKEN_IDENTIFIER: {
				string ident = (*i)->identifier();
				// search in local data
				if (ctx.is_declared(ident)) break; // nothing special, just local variable or parameter
				bool found = false;
				// search in static data
				for (DSet::const_iterator j = sdata.begin(); !found && j != sdata.end(); ++j)
					if ((*j).get_name() == ident && (*j).get_module() == (*i)->get_module())
					{
						collect_unique_data_access(*j);
						found = true;
					}
				// search in global data
				for (DSet::const_iterator j = gdata.begin(); !found && j != gdata.end(); ++j)
					if ((*j).get_name() == ident)
					{
						collect_unique_data_access(*j);
						found = true;
					}
				} break;
		}
		collect_data_recursive_depth(*i, ctx);
	}
}

std::string AnalysisReport::get_identifier_of_parameter(const Node * parameter) const
{
	if (!parameter) return string("");
	if (!parameter->is_type(Parser::TOKEN_PARAMETER)) return string("");
	for (Node::const_iterator i = parameter->begin(); i != parameter->end(); ++i)
		if ((*i)->is_type(Parser::TOKEN_IDENTIFIER)) return (*i)->identifier();
	return string("");
}

void AnalysisReport::collect_function_parameters(const Node * func, DataCollectorContext & ctx)
{
	if (!func) return;
	if (!func->is_type(Parser::TOKEN_FUNCTION_DEFINITION)) return;
	Node * direct_declarator = 0;
	for (Node::const_iterator i = func->begin(); !direct_declarator && i != func->end(); ++i)
		if ((*i)->is_type(Parser::TOKEN_DIRECT_DECLARATOR)) direct_declarator = *i;
	if (!direct_declarator) return;
	for (Node::const_iterator i = direct_declarator->begin(); i != direct_declarator->end(); ++i)
	{
		switch ((*i)->get_type())
		{
			case Parser::TOKEN_PARAMETER_LIST: // two or more parameters
				for (Node::const_iterator j = (*i)->begin(); j != (*i)->end(); ++j)
				{ // ignore Parser::TOKEN_COMMA
					if ((*j)->is_type(Parser::TOKEN_PARAMETER)) ctx.add(get_identifier_of_parameter(*j));
				}
				return;
			case Parser::TOKEN_PARAMETER: // just one parameter
				ctx.add(get_identifier_of_parameter(*i));
				return;
		}
	}
}

void AnalysisReport::collect_data_accesses_from_func(const Node * func)
{	// 'caller' is defined and ready
	DataCollectorContext ctx;
	collect_function_parameters(func, ctx);
	for (Node::const_iterator i = func->begin(); i != func->end(); ++i)
		if ((*i)->is_type(Parser::TOKEN_COMPOUND_STATEMENT))
		{
			collect_data_recursive_depth(*i, ctx);
			return;
		}
}

void AnalysisReport::collect_data_from_declaration(const Node * decl)
{
	DSet * col = &gdata; // default: global data
	const Node * sub = (*decl)[0];
	if (sub->is_type(Parser::TOKEN_DECLARATION_SPECIFIERS))
	{
		for (Node::const_iterator i = sub->begin(); i != sub->end(); ++i)
		{
			switch ((*i)->get_type())
			{
				case Parser::TOKEN_EXTERN:  return; // external symbols are not of interest
				case Parser::TOKEN_TYPEDEF: return; // typedefs are not of interest
				case Parser::TOKEN_CONST:   return; // constants are not of interest
				case Parser::TOKEN_STATIC:  col = &sdata; break; // static data
			}
		}
	}
	sub = (*decl)[1];
	string module = decl->get_module();
	string s;
	const Node * snode = 0;
	switch (sub->get_type())
	{
		case Parser::TOKEN_IDENTIFIER:
			col->insert(Declaration(sub->identifier(), module));
			break;
		case Parser::TOKEN_INIT_DECLARATOR:
			snode = (*sub)[0];
			switch (snode->get_type())
			{
				case Parser::TOKEN_IDENTIFIER:
					col->insert(Declaration(snode->identifier(), module));
					break;
				case Parser::TOKEN_DIRECT_DECLARATOR:
					if (extract_identifier_of_direct_declarator(snode, s))
						col->insert(Declaration(s, module));
					break;
				case Parser::TOKEN_DECLARATOR:
					for (Node::const_iterator i = snode->begin(); i != snode->end(); ++i)
						if ((*i)->is_type(Parser::TOKEN_IDENTIFIER))
							col->insert(Declaration((*i)->identifier(), module));
					break;
				default:
					cerr << __FUNCTION__ << ":" << __LINE__ << ": ERROR: UNKNOWN TOKEN" << endl;
					break;
			}
			break;
		case Parser::TOKEN_DECLARATOR:
			for (Node::const_iterator i = sub->begin(); i != sub->end(); ++i)
				if ((*i)->is_type(Parser::TOKEN_IDENTIFIER))
					col->insert(Declaration((*i)->identifier(), module));
			break;
		case Parser::TOKEN_INIT_DECLARATOR_LIST:
			for (Node::const_iterator i = sub->begin(); i != sub->end(); ++i)
				if ((*i)->is_type(Parser::TOKEN_IDENTIFIER))
					col->insert(Declaration((*i)->identifier(), module));
			break;
		case Parser::TOKEN_DIRECT_DECLARATOR:
			if (extract_identifier_of_direct_declarator(sub, s))
				col->insert(Declaration(s, module));
			break;
	}
}

void AnalysisReport::collect_data(const STVec & trees)
{
	for (STVec::const_iterator tree = trees.begin(); tree != trees.end(); ++tree)
	{
		for (SyntaxTree::const_iterator i = (*tree)->begin(); i != (*tree)->end(); ++i)
		{
			if ((*i)->is_type(Parser::TOKEN_DECLARATION)) collect_data_from_declaration(*i);
		}
	}
}

bool AnalysisReport::extract_identifier_of_direct_declarator(const Node * node, std::string & ident) const
{
	for (size_t i = 0; i < node->size(); ++i)
	{
		if ((*node)[i]->is_type(Parser::TOKEN_IDENTIFIER))
		{
			if (i < node->size()-1)
			{
				// could be a function declaration as well!
				if (!(*node)[i+1]->is_type(Parser::TOKEN_PARENTHESIS_LEFT))
				{
					ident = (*node)[i]->identifier();
					return true;
				} return false; // it is a function, so quit!
			}
			else
			{
				ident = (*node)[i]->identifier();
				return true;
			}
		}
		if (extract_identifier_of_direct_declarator((*node)[i], ident)) return true;		
	}
	return false;
}
