#include <iostream>
#include <fstream>
#include <iomanip>
#include <set>
#include <map>
#include <vector>
#include <getopt.h>

using namespace std;

#include "SyntaxTree.h"
#include "Declaration.h"
#include "C99_STD_Parser.h"
#include "C99_Tasking64_Parser.h"
#include "TreeReport.h"
#include "AnalysisReport.h"
#include "version.h"

typedef multimap<string,string,less<string> > SSMMap;
typedef set<string,less<string> > SSet;
typedef vector<string> SVec;

typedef SyntaxTree::STVec STVec;

string prog_name = "";
string output_filename = "";
string error_filename = "";
ofstream output_stream;
ofstream error_stream;

bool verbose = false;
Parser * parser = 0;
Report * report = 0;

// ----------------------------------------------------------------------------
void init_parser(const string & name)
{
	if (parser) delete parser;
	if      (name == "std")   parser = new C99_STD_Parser;
	else if (name == "tsk64") parser = new C99_Tasking64_Parser;
	else                      parser = new C99_STD_Parser;
}

void init_report(const string & name)
{
	if (report) delete report;
	if      (name == "tree")     report = new TreeReport;
	else if (name == "analysis") report = new AnalysisReport;
	else                         report = new TreeReport;
}

static const struct option options[] =
{
	{ "help",    no_argument,       0, 'h' },
	{ "error",   required_argument, 0, 'e' },
	{ "output",  required_argument, 0, 'o' },
	{ "parser",  required_argument, 0, 'p' },
	{ "report",  required_argument, 0, 'r' },
	{ "verbose", no_argument,       0, 'v' },
	{ 0,         0,                 0, 0   }
};
static const char * option_string = "e:ho:p:r:v";

void usage(void)
{
	cerr << endl;
	cerr << "C99 Parser " << VERSION << endl;
	cerr << "usage: " << prog_name
			<< " [-h|--help]"
			<< " [-e|--error filename]"
			<< " [-o|--output filename]"
			<< " [-p|--parser parser]"
			<< " [-r|--report]"
			<< " [-v|--verbose]"
			<< " file(s)" << endl;
	cerr << "   -h          : help" << endl;
	cerr << "   -e filename : send error output to file instead to stderr" << endl;
	cerr << "   -o filename : output of report is written to the specified file" << endl;
	cerr << "                 if this parameter is not specified, the output is" << endl;
	cerr << "                 written to stdout" << endl;
	cerr << "   -p parser   : specification of the parse" << endl;
	cerr << "                 std      : Standard C99 (Default)" << endl;
	cerr << "                 tsk64    : Tasking v6.0 r4" << endl;
	cerr << "   -r report   : specification of the report generator" << endl;
	cerr << "                 tree     : Syntax Tree (Default)" << endl;
	cerr << "                 analysis : Analysis Tool" << endl;
	cerr << "   -v          : verbose" << endl;
	cerr << "   file(s)     : preprocessed source files" << endl;
	cerr << endl;
}

void parse_args(int argc, char ** argv, SVec & filenames)
{
	while (optind < argc)
	{
		int index = -1;
		int result = getopt_long(argc, argv, option_string, options, &index);
		if (result == -1) break;
		switch (result)
		{
			case 'h': usage(); exit(-1); break;
			case 'p': init_parser(optarg); break;
			case 'v': verbose = true; break;
			case 'r': init_report(optarg); break;
			case 'o': output_filename = optarg; break;
			case 'e': error_filename = optarg; break;
			case 0:
				cerr << prog_name << ": error, unknown parameter" << endl;
				exit(-1);
				break;
			default:
				cerr << prog_name << ": error, unknown parameter '" << static_cast<char>(result) << "'" << endl;
				exit(-1);
				break;
		}
	}
	while (optind < argc) filenames.push_back(argv[optind++]);
}

void process_files(SVec & filenames, STVec & trees, Parser * parser)
{
	if (filenames.size() <= 0) return;
	for (SVec::const_iterator it = filenames.begin(); it != filenames.end(); ++it)
	{
		if (verbose) cout << prog_name << ": processing file: '" << *it << "'" << endl;
		ifstream ifs((*it).c_str(), ios::in);
		if (!ifs)
		{
			cerr << prog_name << ": error: cannot open file '" << *it << "'" << endl;
			continue;
		}
		SyntaxTree * tree = new SyntaxTree;
		int result = parser->parse(ifs, *it, tree);
		if (result == 0)
		{
			trees.push_back(tree);
		}
		else
		{
			delete tree;
			cerr << prog_name << ": error during parse process of '" << *it << "'" << endl;
		}
		ifs.close();
	}
	if (verbose) cout << prog_name << ": all files processed..." << endl;
}

void redirect_stdout(void)
{
	if (output_filename.size() > 0)
	{
		output_stream.open(output_filename.c_str(), ios::out | ios::trunc);
		if (!output_stream)
		{
			cerr << prog_name << ": error: cannot open output file: '" << output_filename << "'" << endl;
			exit(-1);
		}
		cout.rdbuf(output_stream.rdbuf());
	}
}

void redirect_stderr(void)
{
	if (error_filename.size() > 0)
	{
		error_stream.open(error_filename.c_str(), ios::out | ios::trunc);
		if (!error_stream)
		{
			cerr << prog_name << ": error: cannot open error file: '" << error_filename << "'" << endl;
			exit(-1);
		}
		cerr.rdbuf(error_stream.rdbuf());
	}
}

void cleanup(void)
{
	if (parser) delete parser;
	if (report) delete report;
	if (Node::get_num_objects() != 0)
	{
		cerr << prog_name << ": error: number of active nodes: " << Node::get_num_objects() << endl;
	}
}

int main(int argc, char ** argv)
{
	atexit(cleanup);
	prog_name = argv[0];
	prog_name = prog_name.substr(prog_name.find_last_of("/\\")+1);
	if (argc < 2)
	{
		usage();
		exit(-1);
	}
	SVec filenames;
	parse_args(argc, argv, filenames);
	if (filenames.size() <= 0)
	{
		cerr << prog_name << ": error no input files specified" << endl;
		usage();
		exit(-1);
	}

	redirect_stdout();
	redirect_stderr();

	if (!parser) parser = new C99_STD_Parser;
	if (!report) report = new TreeReport;

	STVec trees;
	process_files(filenames, trees, parser);
	report->print(trees, cout);

	// clean up
	while (!trees.empty())
	{
		SyntaxTree * tree = trees.back();
		trees.pop_back();
		delete tree;
	}

	return 0;
}

